@@ -4,15 +4,16 @@ import sys
 import py
 import argparse
 import subprocess
-import devpi
 from base64 import b64encode
+from contextlib import closing
+from devpi import hookspecs
 from devpi_common.types import lazydecorator, cached_property
 from devpi_common.url import URL
 from devpi_common.proc import check_output
-from devpi.use import Current
+from devpi.use import PersistentCurrent
 from devpi_common.request import new_requests_session
 from devpi import __version__ as client_version
-
+from pluggy import PluginManager
 import json
 std = py.std
 subcommand = lazydecorator()
@@ -27,7 +28,8 @@ def main(argv=None):
     if argv is None:
         argv = list(sys.argv)
     hub, method = initmain(argv)
-    return method(hub, hub.args)
+    with closing(hub):
+        return method(hub, hub.args)
 
 def initmain(argv):
     args = parse_args(argv)
@@ -38,8 +40,19 @@ def initmain(argv):
     mod = __import__(mod, None, None, ["__doc__"])
     return Hub(args), getattr(mod, func)
 
+
+def get_pluginmanager(load_entry_points=True):
+    pm = PluginManager("devpiclient", implprefix="devpiclient_")
+    pm.add_hookspecs(hookspecs)
+    if load_entry_points:
+        pm.load_setuptools_entrypoints("devpi_client")
+    pm.check_pending()
+    return pm
+
+
 notset = object()
 
+
 class Hub:
     class Popen(std.subprocess.Popen):
         STDOUT = std.subprocess.STDOUT
@@ -55,6 +68,14 @@ class Hub:
         self.quiet = False
         self._last_http_stati = []
         self.http = new_requests_session(agent=("client", client_version))
+        self.pm = get_pluginmanager()
+
+    @property
+    def hook(self):
+        return self.pm.hook
+
+    def close(self):
+        self.http.close()
 
     @property
     def _last_http_status(self):
@@ -75,8 +96,9 @@ class Hub:
 
     # remote http hooks
 
-    def http_api(self, method, url, kvdict=None, quiet=False, auth=notset,
-        check_version=True, fatal=True, type=None):
+    def http_api(self, method, url, kvdict=None, quiet=False,
+                 auth=notset, basic_auth=notset, cert=notset,
+                 check_version=True, fatal=True, type=None, verify=None):
         """ send a json request and return a HTTPReply object which
         adds some extra methods to the requests's Reply object.
 
@@ -89,6 +111,7 @@ class Hub:
         If type is specified and the json result type does not match,
         bail out fatally (unless fatal = False)
         """
+        assert kvdict is None or isinstance(kvdict, dict)
         if isinstance(url, URL):
             url = url.url
         jsontype = "application/json"
@@ -98,10 +121,19 @@ class Hub:
             if auth is notset:
                 auth = self.current.get_auth()
             set_devpi_auth_header(headers, auth)
-            basic_auth = self.current.get_basic_auth(url=url)
-            cert = self.current.get_client_cert(url=url)
+            if basic_auth is notset:
+                basic_auth = self.current.get_basic_auth(url=url)
+            if cert is notset:
+                cert = self.current.get_client_cert(url=url)
             r = self.http.request(method, url, data=data, headers=headers,
-                                  auth=basic_auth, cert=cert)
+                                  auth=basic_auth, cert=cert, verify=verify)
+        except self.http.SSLError as e:
+            # If verify was set, re-raise this so it can be handled and retried as appropriate
+            self._last_http_stati.append(-1)
+            if verify is not None:
+                raise
+            else:
+                self.fatal("SSL verification failed %r:\n%s" % (url, e))
         except self.http.Errors as e:
             self._last_http_stati.append(-1)
             self.fatal("could not connect to %r:\n%s" % (url, e))
@@ -177,16 +209,7 @@ class Hub:
     def current(self):
         self.clientdir.ensure(dir=1)
         path = self.clientdir.join("current.json")
-        return Current(path)
-
-    @property
-    def remoteindex(self):
-        try:
-            return self._remoteindex
-        except AttributeError:
-            from devpi.remoteindex import RemoteIndex
-            self._remoteindex = RemoteIndex(self.current)
-            return self._remoteindex
+        return PersistentCurrent(path)
 
     def get_existing_file(self, arg):
         p = py.path.local(arg, expanduser=True)
@@ -197,12 +220,47 @@ class Hub:
         return p
 
     @property
-    def path_venvbase(self):
+    def venv(self):
+        venvdir = None
+        vbin = "Scripts" if sys.platform == "win32" else "bin"
+
+        if self.args.venv == "-":
+            self.current.reconfigure(dict(venvdir=None))
+        else:
+            if self.args.venv:
+                venvname = self.args.venv
+                cand = self.cwd.join(venvname, vbin, abs=True)
+                if not cand.check() and self.venvwrapper_home:
+                    cand = self.venvwrapper_home.join(venvname, vbin, abs=True)
+                if not cand.check():
+                    self.create_virtualenv(venvname)
+                venvdir = cand.dirpath().strpath
+                self.current.reconfigure(dict(venvdir=venvdir))
+            else:
+                venvdir = self.current.venvdir or self.active_venv()
+
+        return venvdir
+
+    @property
+    def venvwrapper_home(self):
         path = os.environ.get("WORKON_HOME", None)
         if path is None:
             return
         return py.path.local(path)
 
+    def active_venv(self):
+        """current activated virtualenv"""
+        path = os.environ.get("VIRTUAL_ENV", None)
+        if path is None:
+            return
+        return py.path.local(path)
+
+    def create_virtualenv(self, venv):
+        if self.venvwrapper_home:
+            self.popen_check(["mkvirtualenv", venv])
+        else:
+            self.popen_check(["virtualenv", "-q", venv])
+
     def popen_output(self, args, cwd=None, report=True):
         if isinstance(args, str):
             args = std.shlex.split(args)
@@ -218,7 +276,7 @@ class Hub:
             self.report_popen(args, cwd)
         return check_output(args, cwd=str(cwd))
 
-    def popen(self, args, cwd=None):
+    def popen(self, args, cwd=None, dryrun=None, **popen_kwargs):
         if isinstance(args, str):
             args = std.shlex.split(args)
         assert args[0], args
@@ -226,13 +284,16 @@ class Hub:
         if cwd == None:
             cwd = self.cwd
         self.report_popen(args, cwd)
-        if self.args.dryrun:
+        if dryrun is None:
+            dryrun = self.args.dryrun
+        if dryrun:
             return
-        popen = subprocess.Popen(args, cwd=str(cwd))
+        popen = subprocess.Popen(args, cwd=str(cwd), **popen_kwargs)
         out, err = popen.communicate()
         ret = popen.wait()
         if ret:
             self.fatal("****** process returned %s" % ret)
+        return (ret, out, err)
 
     def report_popen(self, args, cwd=None, extraenv=None):
         base = cwd or self.cwd
@@ -244,7 +305,7 @@ class Hub:
                 ["%s=%s" % item for item in extraenv.items()])
         else:
             envadd = ""
-        self.line("--> $", rel, " ".join(args[1:]), envadd)
+        self.line("--> ", base + "$", rel, " ".join(args[1:]), envadd)
 
     def popen_check(self, args, extraenv=None):
         assert args[0], args
@@ -309,6 +370,10 @@ class HTTPReply(object):
     def __getattr__(self, name):
         return getattr(self._response, name)
 
+    @property
+    def reason(self):
+        return self._response.reason
+
     @property
     def type(self):
         return self._json.get("type")
@@ -366,13 +431,41 @@ def try_argcomplete(parser):
         else:
             argcomplete.autocomplete(parser)
 
+
+def print_version(hub):
+    hub.line("devpi-client %s\n" % client_version)
+    if hub.current.rooturl is not None:
+        url = URL(hub.current.rooturl).addpath('+status').url
+        try:
+            r = HTTPReply(hub.http.get(
+                url, headers=dict(accept='application/json')))
+        except hub.http.Errors:
+            pass
+        else:
+            status = r.json_get('result')
+            if r.status_code == 200 and status is not None:
+                hub.info(
+                    "current devpi server: %s" % hub.current.rooturl)
+                versioninfo = status.get('versioninfo', {})
+                for name, version in sorted(versioninfo.items()):
+                    hub.line("    %s %s" % (name, version))
+
+
 def parse_args(argv):
     argv = [str(x) for x in argv]
-    parser = getbaseparser(argv[0])
+    parser = getbaseparser('devpi')
     add_subparsers(parser)
     try_argcomplete(parser)
     try:
-        return parser.parse_args(argv[1:])
+        args = parser.parse_args(argv[1:])
+        if sys.version_info >= (3,) and args.version:
+            with closing(Hub(args)) as hub:
+                print_version(hub)
+            parser.exit()
+        if args.command is None:
+            raise parser.ArgumentError(
+                "the following arguments are required: command")
+        return args
     except parser.ArgumentError as e:
         if not argv[1:]:
             return parser.parse_args(["-h"])
@@ -392,7 +485,7 @@ def add_subparsers(parser):
     subparsers = parser.add_subparsers()
     # see http://stackoverflow.com/questions/18282403/
     # for the following two lines (py3 compat)
-    subparsers.required = True
+    subparsers.required = False
     subparsers.dest = "command"
 
     for func, args, kwargs in subcommand.discover(globals()):
@@ -409,32 +502,43 @@ def add_subparsers(parser):
         func(subparser)
         mainloc = args[0]
         subparser.set_defaults(mainloc=mainloc)
-    #subparser = subparsers.add_parser("_test", help=argparse.SUPPRESS)
-    #subparser.set_defaults(mainloc="devpi")
+
 
 def getbaseparser(prog):
     parser = MyArgumentParser(prog=prog, description=main_description)
-    add_generic_options(parser)
+    add_generic_options(parser, defaults=True)
     return parser
 
-def add_generic_options(parser):
+def add_generic_options(parser, defaults=False):
     group = parser.add_argument_group("generic options")
-    group.add_argument("--version", action="version",
-                       version=devpi.__version__)
+    if sys.version_info < (3,):
+        # workaround old argparse which doesn't support optional sub commands
+        group.add_argument("--version", action="version",
+            version="devpi-client %s" % client_version)
+    else:
+        group.add_argument("--version", action="store_true",
+            help="show program's version number and exit")
     group.add_argument("--debug", action="store_true",
         help="show debug messages including more info on server requests")
     group.add_argument("-y", action="store_true", dest="yes",
         help="assume 'yes' on confirmation questions")
     group.add_argument("-v", "--verbose", action="count",
         help="increase verbosity")
+    # workaround for http://bugs.python.org/issue23058
+    if defaults:
+        clientdir_default = os.path.expanduser(
+            os.environ.get("DEVPI_CLIENTDIR", "~/.devpi/client"))
+    else:
+        # subcommands will have their default being suppressed, so only the
+        # main one is used
+        clientdir_default = argparse.SUPPRESS
     group.add_argument("--clientdir", action="store", metavar="DIR",
-        default=os.path.expanduser(os.environ.get("DEVPI_CLIENTDIR",
-                                                  "~/.devpi/client")),
+        default=clientdir_default,
         help="directory for storing login and other state")
 
 @subcommand("devpi.quickstart")
 def quickstart(parser):
-    """ start a server, create a user and login, then create a USER/dev
+    """ (deprecated) start a server, create a user and login, then create a USER/dev
     index and then connect to this index, so that subsequent devpi
     commands can work with it.
     """
@@ -464,8 +568,15 @@ def use(parser):
 
     parser.add_argument("--set-cfg", action="store_true", default=None,
         dest="setcfg",
-        help="create or modify pip/setuptools config files in home directory "
-             "so pip/easy_install will pick up the current devpi index url")
+        help="create or modify pip/setuptools config files so "
+             "pip/easy_install will pick up the current devpi index url. "
+             "If a virtualenv is activated, only its pip config will be set.")
+    parser.add_argument("-t", "--pip-set-trusted", choices=["yes", "no", "auto"], default="auto",
+        dest="settrusted",
+        help="when used in conjunction with set-cfg, also set matching "
+             "pip trusted-host setting for the provided devpi index url. "
+             "With 'auto', trusted will be set for http urls or hosts that "
+             "fail https ssl validation. 'no' will clear setting")
     parser.add_argument("--always-set-cfg",
         choices=["yes", "no"], default=None,
         dest="always_setcfg",
@@ -474,7 +585,9 @@ def use(parser):
              "config file and can be cleared with '--always-set-cfg=no'.")
     parser.add_argument("--venv", action="store", default=None,
         help="set virtual environment to use for install activities. "
-             "specify '-' to unset it.")
+             "specify '-' to unset it. "
+             "venv be created if given name doesn't already exist. "
+             "Note: an activated virtualenv will be used without needing this.")
     parser.add_argument("--urls", action="store_true",
         help="show remote endpoint urls")
     parser.add_argument("-l", action="store_true", dest="list",
@@ -501,8 +614,8 @@ def getjson(parser):
     A low-level command to show json-formatted configuration data
     from remote resources.  This will always query the remote server.
     """
-    parser.add_argument("path", action="store",
-        help="path to a resource to show information on. "
+    parser.add_argument("path", action="store", metavar="path_or_url",
+        help="path or url of a resource to show information on. "
              "examples: '/', '/user', '/user/index'.")
 
 @subcommand("devpi.getjson:main_patchjson")
@@ -536,6 +649,9 @@ def list_(parser):
     parser.add_argument("--all", action="store_true",
         help="show all versions instead of just the newest")
 
+    parser.add_argument("--index", default=None,
+        help="index to look at (defaults to current index)")
+
     parser.add_argument("spec", nargs="?",
         help="show info for a project or a specific release. "
              "Example specs: pytest or 'pytest>=2.3.5'"
@@ -549,6 +665,8 @@ def remove(parser):
     current index (see "devpi use").  It will ask interactively
     for confirmation before performing the actual removals.
     """
+    parser.add_argument("--index", default=None,
+        help="index to remove from (defaults to current index)")
     parser.add_argument("spec",
         help="remove info/files for a project/version/release file from the "
              "current index. "
@@ -564,13 +682,14 @@ def user(parser):
     If you create a user you either need to pass a ``password=...``
     setting or interactively type a password.
     """
-    parser.add_argument("-c", "--create", action="store_true",
+    group = parser.add_mutually_exclusive_group()
+    group.add_argument("-c", "--create", action="store_true",
         help="create a user")
-    parser.add_argument("--delete", action="store_true",
+    group.add_argument("--delete", action="store_true",
         help="delete a user")
-    parser.add_argument("-m", "--modify", action="store_true",
+    group.add_argument("-m", "--modify", action="store_true",
         help="modify user settings")
-    parser.add_argument("-l", "--list", action="store_true",
+    group.add_argument("-l", "--list", action="store_true",
         help="list user names")
     parser.add_argument("username", type=str, action="store", nargs="?",
         help="user name")
@@ -578,6 +697,14 @@ def user(parser):
         help="key=value configuration item.  Possible keys: "
              "email, password.")
 
+
+@subcommand("devpi.user:passwd")
+def passwd(parser):
+    """ change password of specified user or current user if not specified. """
+    parser.add_argument(
+        "username", type=str, action="store", nargs="?", help="user name")
+
+
 @subcommand("devpi.login")
 def login(parser):
     """ login to devpi-server with the specified user.
@@ -598,6 +725,15 @@ def logoff(parser):
     This will erase the client-side login token (see "devpi login").
     """
 
+
+@subcommand("devpi.login:logoff")
+def logout(parser):
+    """ log out of the current devpi-server.
+
+    This will erase the client-side login token (see "devpi login").
+    """
+
+
 @subcommand("devpi.index")
 def index(parser):
     """ create, delete and manage indexes.
@@ -609,6 +745,9 @@ def index(parser):
     You can also view the configuration of any index with
     ``devpi index USER/INDEX`` or list all indexes of the
     in-use index with ``devpi index -l``.
+
+    For the ``keyvalues`` option CSV means "Comma Separated Value", in other
+    words, a list of values separated by commas.
     """
     group = parser.add_mutually_exclusive_group()
     group.add_argument("-c", "--create", action="store_true",
@@ -624,11 +763,6 @@ def index(parser):
         help="key=value configuration item. Possible key=value are "
              "bases=CSV, volatile=True|False, acl_upload=CSV)")
 
-if sys.platform == "win32":
-    default_format = "sdist.zip"
-else:
-    default_format = "sdist.tgz"
-
 @subcommand("devpi.upload")
 def upload(parser):
     """ (build and) upload packages to the current devpi-server index.
@@ -640,6 +774,10 @@ def upload(parser):
 
     Or, if you don't specify any path, a setup.py file must exist
     and will be used to perform build and upload commands.
+
+    If you have a ``setup.cfg`` file you can have a "[devpi:upload]" section
+    with ``formats``, ``no-vcs = 1``, and ``setupdir-only = 1`` settings
+    providing defaults for the respective command line options.
     """
     #parser.add_argument("--ver", dest="setversion",
     #    action="store", default=None,
@@ -653,8 +791,11 @@ def upload(parser):
              "directly using their dirname as current dir. By default "
              "git/hg/svn/bazaar are auto-detected and packaging is run from "
              "a fresh directory with all versioned files exported.")
+    build.add_argument("--setupdir-only", action="store_true",
+        dest="setupdironly",
+        help="VCS-export only the directory containing setup.py")
 
-    build.add_argument("--formats", default=default_format, action="store",
+    build.add_argument("--formats", default="", action="store",
         help="comma separated list of build formats (passed to setup.py). "
              "Examples sdist.zip,bdist_egg,bdist_wheel,bdist_dumb.")
     build.add_argument("--with-docs", action="store_true", default=None,
@@ -666,6 +807,8 @@ def upload(parser):
         help="as --with-docs but don't build or upload release files")
 
     direct = parser.add_argument_group("direct file upload options")
+    direct.add_argument("--index", default=None,
+        help="index to upload to (defaults to current index)")
     direct.add_argument("--from-dir", action="store_true", default=None,
         dest="fromdir",
         help="recursively look for archive files in path if it is a dir")
@@ -707,8 +850,35 @@ def test(parser):
         help="extra command line arguments for tox. e.g. "
              "--toxargs=\"-c othertox.ini\"")
 
+    parser.add_argument("--detox", "-d", action="store_true",
+        dest="detox", default=False,
+        help="(experimental) run tests concurrently in multiple processes using "
+             "the detox tool (which must be installed)")
+
+    parser.add_argument("--no-upload", action="store_false",
+        dest="upload_tox_results", default=True,
+        help="Skip upload of tox results")
+
+    parser.add_argument("--index", default=None,
+        help="index to get package from, defaults to current index. "
+             "Either just the NAME, using the current user, USER/NAME using "
+             "the current server or a full URL for another server.")
+
+    parser.add_argument(
+        "--select", "-s", metavar="SELECTOR",
+        type=str, default=None, action="store",
+        help="Selector for release files. "
+             "This is a regular expression to select release files for which "
+             "tests will be run. With this option it's possible to select "
+             "wheels that aren't universal, or run tests only for one "
+             "specific release file.")
+
+    parser.add_argument(
+        "--list", "-l", action="store_true",
+        help="Just list the release files which would be tested.")
+
     parser.add_argument("pkgspec", metavar="pkgspec", type=str,
-        default=None, action="store", nargs=1,
+        default=None, action="store", nargs="+",
         help="package specification in pip/setuptools requirement-syntax, "
              "e.g. 'pytest' or 'pytest==2.4.2'")
 
@@ -722,12 +892,14 @@ def push(parser):
         needs to be defined in your ``.pypirc`` file.  Or you can
         push to another devpi index ("user/name").
     """
+    parser.add_argument("--index", default=None,
+        help="index to push from (defaults to current index)")
     parser.add_argument("--pypirc", metavar="path", type=str,
         default=None, action="store",
         help="path to pypirc")
-    parser.add_argument("nameversion", metavar="NAME-VER", type=str,
+    parser.add_argument("pkgspec", metavar="pkgspec", type=str,
         default=None, action="store",
-        help="release in format 'name-version'. of which the metadata and "
+        help="release in format 'name==version'. of which the metadata and "
              "all release files are to be uploaded to the specified "
              "external pypi repo." )
     parser.add_argument("target", metavar="TARGETSPEC", type=str,
@@ -745,6 +917,8 @@ def install(parser):
     This is convenience wrapper which configures and invokes
     ``pip install`` commands for you, using the current index.
     """
+    parser.add_argument("--index", default=None,
+        help="index to get package from (defaults to current index)")
     parser.add_argument("-l", action="store_true", dest="listinstalled",
         help="print list of currently installed packages. ")
     parser.add_argument("-e", action="store", dest="editable", metavar="ARG",
@@ -752,11 +926,28 @@ def install(parser):
     parser.add_argument("--venv", action="store", metavar="DIR",
         help="install into specified virtualenv (created on the fly "
              "if none exists).")
+    parser.add_argument("-r", "--requirement", action="store_true",
+        help="Install from the given requirements file.")
     parser.add_argument("pkgspecs", metavar="pkg", type=str,
         action="store", default=None, nargs="*",
         help="uri or package file for installation from current index. """
     )
 
+
+@subcommand("devpi.refresh")
+def refresh(parser):
+    """ invalidates the mirror caches for the specified package(s).
+
+    In case your devpi server hasn't updated the list of latest releases, this
+    forces a reload of the them (EXPERIMENTAL).
+    """
+    parser.add_argument("--index", default=None,
+        help="index to refresh (defaults to current index)")
+    parser.add_argument(
+        "pkgnames", metavar="pkg", type=str, action="store", nargs="+",
+        help="package name to refresh.""")
+
+
 def verify_reply_version(hub, reply):
     acceptable_api_version = ("2",)
     version = reply.headers.get("X-DEVPI-API-VERSION", None)
