@@ -1,5 +1,5 @@
 import py
-from devpi_common.metadata import splitbasename
+from devpi_common.metadata import parse_requirement, splitbasename
 from . import pypirc
 
 class PyPIPush:
@@ -12,18 +12,20 @@ class PyPIPush:
         req = dict(name=name, version=str(version), posturl=self.posturl,
                    username=self.user, password=self.password )
         index = hub.current.index
-        return hub.http_api("push", index, kvdict=req, fatal=False)
+        return hub.http_api("push", index, kvdict=req, fatal=True)
 
         #assert r.status_code == 200, r.content
 
+
 class DevpiPush:
-    def __init__(self, targetindex):
+    def __init__(self, targetindex, index):
         self.targetindex = targetindex
+        self.index = index
 
     def execute(self, hub, name, version):
         req = dict(name=name, version=str(version),
                    targetindex=self.targetindex)
-        return hub.http_api("push", hub.current.index, kvdict=req, fatal=False)
+        return hub.http_api("push", self.index, kvdict=req, fatal=True)
 
 def parse_target(hub, args):
     if args.target.startswith("pypi:"):
@@ -45,11 +47,31 @@ def parse_target(hub, args):
     if args.target.count("/") != 1:
         hub.fatal("target %r not of form USER/NAME or pypi:REPONAME" % (
                   args.target, ))
-    return DevpiPush(args.target)
+    index = hub.current.index
+    if args.index:
+        if args.index.count("/") > 1:
+            hub.fatal("index %r not of form USER/NAME or NAME" % args.index)
+        index = hub.current.get_index_url(args.index, slash=False)
+    return DevpiPush(args.target, index)
+
 
 def main(hub, args):
     pusher = parse_target(hub, args)
-    name, version = splitbasename(args.nameversion + ".zip")[:2]
+    name = None
+    version = None
+    if '==' not in args.pkgspec and '-' in args.pkgspec:
+        name, version = splitbasename(args.pkgspec + ".zip")[:2]
+    if not name or not version:
+        req = parse_requirement(args.pkgspec)
+        if len(req.specs) != 1 or req.specs[0][0] != '==':
+            hub.fatal(
+                "The release specification needs to be of this form: name==version")
+        name = req.project_name
+        version = req.specs[0][1]
+    else:
+        hub.warn(
+            "Old style package specification is deprecated, "
+            "use this form: your-pkg-name==your.version.specifier")
     r = pusher.execute(hub, name, version)
     if r.type == "actionlog":
         for action in r["result"]:
