@@ -1,11 +1,13 @@
 import os
+import sys
 import py
+import re
 
 import check_manifest
 
-from devpi import log
-from devpi_common.metadata import Version, BasenameMeta, get_pyversion_filetype
+from devpi_common.metadata import Version, get_pyversion_filetype
 from devpi_common.archive import zip_dir
+from devpi_common.types import CompareMixin
 from .main import HTTPReply, set_devpi_auth_header
 
 def main(hub, args):
@@ -16,7 +18,7 @@ def main(hub, args):
     # so we need to monkeypatch distutils in the setup.py process
 
     current = hub.require_valid_current_with_index()
-    if not current.pypisubmit:
+    if not args.index and not current.pypisubmit:
         hub.fatal("no pypisubmit endpoint available for: %s" % current.index)
 
     if args.path:
@@ -25,19 +27,28 @@ def main(hub, args):
     setup = hub.cwd.join("setup.py")
     if not setup.check():
         hub.fatal("no setup.py found in", hub.cwd)
-    checkout = Checkout(hub, hub.cwd)
+
+    setupcfg = read_setupcfg(hub, hub.cwd)
+    checkout = Checkout(hub, hub.cwd, hasvcs=setupcfg.get("no-vcs"),
+                        setupdir_only=setupcfg.get("setupdir-only"))
     uploadbase = hub.getdir("upload")
     exported = checkout.export(uploadbase)
 
     exported.prepare()
     archives = []
     if not args.onlydocs:
-        archives.extend(exported.setup_build())
+        archives.extend(exported.setup_build(
+                            default_formats=setupcfg.get("formats")))
     if args.onlydocs or args.withdocs:
-        archives.append(exported.setup_build_docs())
+        p = exported.setup_build_docs()
+        if p:
+            archives.append(p)
     if not archives:
         hub.fatal("nothing built!")
-    uploader = Uploader(hub, args)
+    name_version = exported.setup_name_and_version()
+    uploader = Uploader(hub, args, name_version=name_version)
+    if args.index:
+        uploader.pypisubmit = hub.current.get_index_url(args.index).url
     uploader.do_upload_paths(archives)
 
 def filter_latest(path_pkginfo):
@@ -64,12 +75,19 @@ def main_fromfiles(hub, args):
         paths.append(p)
 
     uploader = Uploader(hub, args)
+    if args.index:
+        uploader.pypisubmit = hub.current.get_index_url(args.index).url
     uploader.do_upload_paths(paths)
 
 class Uploader:
-    def __init__(self, hub, args):
+    def __init__(self, hub, args, name_version=None):
         self.hub = hub
         self.args = args
+        # allow explicit name and version instead of using pkginfo which
+        # has a high failure rate for documentation zips because they miss
+        # explicit metadata and the implementation has to guess
+        self.name_version = name_version
+        self.pypisubmit = hub.current.pypisubmit
 
     def do_upload_paths(self, paths):
         hub = self.hub
@@ -99,8 +117,12 @@ class Uploader:
 
 
     def upload_doc(self, path, pkginfo):
+        if self.name_version:
+            (name, version) = self.name_version
+        else:
+            (name, version) = (pkginfo.name, pkginfo.version)
         self.post("doc_upload", path,
-                {"name": pkginfo.name, "version": pkginfo.version})
+                {"name": name, "version": version})
 
     def post(self, action, path, meta):
         hub = self.hub
@@ -121,16 +143,17 @@ class Uploader:
         else:
             files = None
         if path:
-            msg = "%s of %s to %s" %(action, path.basename, hub.current.pypisubmit)
+            msg = "%s of %s to %s" %(action, path.basename, self.pypisubmit)
         else:
             msg = "%s %s-%s to %s" %(action, meta["name"], meta["version"],
-                                     hub.current.pypisubmit)
+                                     self.pypisubmit)
         if self.args.dryrun:
             hub.line("skipped: %s" % msg)
         else:
-            r = hub.http.post(hub.current.pypisubmit, dic, files=files,
-                              headers=headers, auth=hub.current.get_basic_auth(),
-                              cert=hub.current.get_client_cert())
+            r = hub.http.post(self.pypisubmit, dic, files=files,
+                              headers=headers,
+                              auth=hub.current.get_basic_auth(self.pypisubmit),
+                              cert=hub.current.get_client_cert(self.pypisubmit))
             hub._last_http_stati.append(r.status_code)
             r = HTTPReply(r)
             if r.status_code == 200:
@@ -141,12 +164,16 @@ class Uploader:
                 if r.type == "actionlog":
                     for x in r.result:
                         hub.error("  " + x)
-                hub.fatal("POST to %s FAILED" % hub.current.pypisubmit)
+                elif r.reason:
+                    hub.error(r.reason)
+                hub.fatal("POST to %s FAILED" % self.pypisubmit)
 
     def upload_release_file(self, path, pkginfo):
         meta = {}
         for attr in pkginfo:
             meta[attr] = getattr(pkginfo, attr)
+        if self.name_version:
+            (meta['name'], meta['version']) = self.name_version
         self.post("register", None, meta=meta)
         pyver = get_pyversion_filetype(path.basename)
         meta["pyversion"], meta["filetype"] = pyver
@@ -165,9 +192,68 @@ def get_archive_files(path):
             if x.basename.endswith(name):
                 yield x
 
+
+# this regexp is taken from pip 8.1.2 (from the vendored packaging)
+VERSION_PATTERN = r"""
+    v?
+    (?:
+        (?:(?P<epoch>[0-9]+)!)?                           # epoch
+        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
+        (?P<pre>                                          # pre-release
+            [-_\.]?
+            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
+            [-_\.]?
+            (?P<pre_n>[0-9]+)?
+        )?
+        (?P<post>                                         # post release
+            (?:-(?P<post_n1>[0-9]+))
+            |
+            (?:
+                [-_\.]?
+                (?P<post_l>post|rev|r)
+                [-_\.]?
+                (?P<post_n2>[0-9]+)?
+            )
+        )?
+        (?P<dev>                                          # dev release
+            [-_\.]?
+            (?P<dev_l>dev)
+            [-_\.]?
+            (?P<dev_n>[0-9]+)?
+        )?
+    )
+    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
+"""
+
+
+name_version_regex = re.compile(
+    "(.*)-(" + VERSION_PATTERN + ")",
+    re.VERBOSE | re.IGNORECASE)
+
+
+def get_name_version_doczip(basename):
+    DOCZIPSUFFIX = ".doc.zip"
+    assert basename.endswith(DOCZIPSUFFIX)
+    fn = basename[:-len(DOCZIPSUFFIX)]
+    name, version = name_version_regex.match(fn).groups()[:2]
+    return name, version
+
+
+class DocZipMeta(CompareMixin):
+    def __init__(self, archivepath):
+        basename = py.path.local(archivepath).basename
+        name, version = get_name_version_doczip(basename)
+        self.name = name
+        self.version = version
+        self.cmpval = (name, version)
+
+    def __repr__(self):
+        return "<DocZipMeta name=%r version=%r>" % (self.name, self.version)
+
+
 def get_pkginfo(archivepath):
     if str(archivepath).endswith(".doc.zip"):
-        return BasenameMeta(archivepath)
+        return DocZipMeta(archivepath)
 
     import pkginfo
     info = pkginfo.get_metadata(str(archivepath))
@@ -184,22 +270,35 @@ def find_parent_subpath(startpath, relpath, raising=True):
 
 
 class Checkout:
-    def __init__(self, hub, setupdir):
+    def __init__(self, hub, setupdir, hasvcs=None, setupdir_only=None):
         self.hub = hub
-        self.rootpath = setupdir
         assert setupdir.join("setup.py").check(), setupdir
-        hasvcs = not hub.args.novcs
+        hasvcs = not hasvcs and not hub.args.novcs
+        setupdir_only = bool(setupdir_only or hub.args.setupdironly)
         if hasvcs:
-            with self.rootpath.as_cwd():
+            with setupdir.as_cwd():
                 try:
-                    hasvcs = check_manifest.detect_vcs().__name__
+                    hasvcs = check_manifest.detect_vcs().metadata_name
                 except check_manifest.Failure:
-                    hasvcs = False
+                    hasvcs = None
+                else:
+                    if hasvcs not in (".hg", ".git") or setupdir_only:
+                        # XXX for e.g. svn we don't do copying
+                        self.rootpath = setupdir
+                    else:
+                        for p in setupdir.parts(reverse=True):
+                            if p.join(hasvcs).exists():
+                                self.rootpath = p
+                                break
+                        else:
+                            hasvcs = None
         self.hasvcs = hasvcs
+        self.setupdir = setupdir
+        self.setupdir_only = setupdir_only
 
     def export(self, basetemp):
         if not self.hasvcs:
-            return Exported(self.hub, self.rootpath, self.rootpath)
+            return Exported(self.hub, self.setupdir, self.setupdir)
         with self.rootpath.as_cwd():
             files = check_manifest.get_vcs_files()
         newrepo = basetemp.join(self.rootpath.basename)
@@ -208,11 +307,21 @@ class Checkout:
             if source.isfile():
                 dest = newrepo.join(fn)
                 dest.dirpath().ensure(dir=1)
-                source.copy(dest)
-        log.debug("copied %s files to %s", len(files), newrepo)
-        self.hub.info("%s-exported project to %s -> new CWD" %(
+                source.copy(dest, mode=True)
+        self.hub.debug("copied", len(files), "files to", newrepo)
+
+        if self.hasvcs not in (".git", ".hg") or self.setupdir_only:
+            self.hub.warn("not copying vcs repository metadata for", self.hasvcs)
+        else:
+            srcrepo = self.rootpath.join(self.hasvcs)
+            assert srcrepo.exists(), srcrepo
+            destrepo = newrepo.join(self.hasvcs)
+            self.rootpath.join(self.hasvcs).copy(destrepo, mode=True)
+            self.hub.info("copied repo", srcrepo, "to", destrepo)
+        self.hub.debug("%s-exported project to %s -> new CWD" %(
                       self.hasvcs, newrepo))
-        return Exported(self.hub, newrepo, self.rootpath)
+        setupdir_newrepo = newrepo.join(self.setupdir.relto(self.rootpath))
+        return Exported(self.hub, setupdir_newrepo, self.setupdir)
 
 class Exported:
     def __init__(self, hub, rootpath, origrepo):
@@ -232,10 +341,12 @@ class Exported:
         setup_py = self.rootpath.join("setup.py")
         if not setup_py.check():
             self.hub.fatal("no setup.py file")
-        name = self.hub.popen_output([self.python, setup_py, "--name"],
-                                     report=False).strip()
+        name = self.hub.popen_output(
+            [self.python, setup_py, "--name"],
+            report=False).splitlines()[-1].strip()
         version = self.hub.popen_output(
-            [self.python, setup_py, "--version"], report=False).strip()
+            [self.python, setup_py, "--version"],
+            report=False).splitlines()[-1].strip()
         self.hub.debug("name, version = %s, %s" %(name, version))
         return name, version
 
@@ -258,8 +369,14 @@ class Exported:
             self.target_distdir.remove()
         self.target_distdir.mkdir()
 
-    def setup_build(self):
-        formats = [x.strip() for x in self.hub.args.formats.split(",")]
+    def setup_build(self, default_formats=None):
+        formats = self.hub.args.formats
+        if not formats:
+            formats = default_formats
+            if not formats:
+                formats = "sdist.zip" if sys.platform == "win32" else "sdist.tgz"
+
+        formats = [x.strip() for x in formats.split(",")]
 
         archives = []
         for format in formats:
@@ -295,9 +412,11 @@ class Exported:
         name, version = self.setup_name_and_version()
         cwd = self.rootpath
         build = cwd.join("build")
-        self.hub.popen_output(
+        out = self.hub.popen_output(
             [self.python, "setup.py", "build_sphinx", "-E",
-             "--build-dir", build])
+             "--build-dir", build], cwd=self.rootpath)
+        if out is None:
+            return
         p = self.target_distdir.join("%s-%s.doc.zip" %(name, version))
         html = build.join("html")
         zip_dir(html, p)
@@ -329,3 +448,11 @@ def sdistformat(format):
         res = format
     return res
 
+
+def read_setupcfg(hub, path):
+    setup_cfg = path.join("setup.cfg")
+    if setup_cfg.exists():
+        cfg = py.iniconfig.IniConfig(setup_cfg)
+        hub.line("detected devpi:upload section in %s" % setup_cfg, bold=True)
+        return cfg.sections.get("devpi:upload", {})
+    return {}
