@@ -1,17 +1,25 @@
-
-# content of conftest.py
+from __future__ import print_function
+from contextlib import closing
+from time import sleep
 import os
-import random
+import pkg_resources
 import pytest
+import socket
 import textwrap
 import py
 import sys
 import json
-from devpi import log
 
 from _pytest.pytester import RunResult, LineMatcher
 from devpi.main import Hub, initmain, parse_args
 from devpi_common.url import URL
+from devpi_server import __version__ as devpi_server_version
+from test_devpi_server.conftest import reqmock  # noqa
+try:
+    from test_devpi_server.conftest import simpypi, simpypiserver  # noqa
+except ImportError:
+    # when testing with older devpi-server
+    pass
 
 import subprocess
 
@@ -27,8 +35,6 @@ def pytest_addoption(parser):
                      action="store", dest="live_url")
 
 
-import subprocess as gsub
-
 def print_info(*args, **kwargs):
     kwargs.setdefault("file", sys.stderr)
     return py.builtin.print_(*args, **kwargs)
@@ -41,14 +47,15 @@ class PopenFactory:
         args = [str(x) for x in args]
         if pipe:
             print ("$ %s [piped]" %(" ".join(args),))
-            popen = gsub.Popen(args, stdout=gsub.PIPE, stderr=gsub.STDOUT)
+            popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
         else:
             showkwargs = " ".join(["%s=%s"] % (x,y) for x,y in kwargs.items())
             print ("$ %s %s" %(" ".join(args), showkwargs))
-            popen = gsub.Popen(args, **kwargs)
+            popen = subprocess.Popen(args, **kwargs)
         def fin():
             try:
                 popen.kill()
+                popen.wait()
             except OSError:
                 print ("could not kill %s" % popen.pid)
         self.addfinalizer(fin)
@@ -102,28 +109,87 @@ def get_pypirc_patcher(devpi):
                 self.saved.copy(self.pypirc)
     return overwrite()
 
-@pytest.fixture(scope="session")
+
+def get_open_port(host):
+    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
+        s.bind((host, 0))
+        s.listen(1)
+        port = s.getsockname()[1]
+    return port
+
+
+def wait_for_port(host, port, timeout=60):
+    while timeout > 0:
+        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
+            s.settimeout(1)
+            if s.connect_ex((host, port)) == 0:
+                return
+        sleep(1)
+        timeout -= 1
+    raise RuntimeError(
+        "The port %s on host %s didn't become accessible" % (port, host))
+
+
+def _liveserver(clientdir):
+    host = 'localhost'
+    port = get_open_port(host)
+    path = py.path.local.sysfind("devpi-server")
+    assert path
+    try:
+        args = [
+            str(path), "--serverdir", str(clientdir), "--debug",
+            "--host", host, "--port", str(port)]
+        server_version = pkg_resources.parse_version(devpi_server_version)
+        if server_version >= pkg_resources.parse_version('4.2.0.dev'):
+            subprocess.check_call(args + ['--init'])
+    except subprocess.CalledProcessError as e:
+        # this won't output anything on Windows
+        print(
+            getattr(e, 'output', "Can't get process output on Windows"),
+            file=sys.stderr)
+        raise
+    p = subprocess.Popen(args)
+    wait_for_port(host, port)
+    return (p, URL("http://%s:%s" % (host, port)))
+
+
+@pytest.yield_fixture(scope="session")
 def url_of_liveserver(request):
     if request.config.option.fast:
         pytest.skip("not running functional tests in --fast mode")
     if request.config.option.live_url:
-        return URL(request.config.option.live_url)
-    port = random.randint(2001, 64000)
+        yield URL(request.config.option.live_url)
+        return
     clientdir = request.config._tmpdirhandler.mktemp("liveserver")
-    path = py.path.local.sysfind("devpi-server")
-    assert path
-    subprocess.check_call([str(path), "--serverdir", str(clientdir),
-                           "--debug",
-                             "--port", str(port), "--start"])
-    def stop():
-        subprocess.check_call(["devpi-server", "--serverdir", str(clientdir),
-                                 "--stop"])
-    request.addfinalizer(stop)
-    return URL("http://localhost:%s" % port)
+    (p, url) = _liveserver(clientdir)
+    try:
+        yield url
+    finally:
+        p.terminate()
+        p.wait()
+
+
+@pytest.yield_fixture(scope="session")
+def url_of_liveserver2(request):
+    if request.config.option.fast:
+        pytest.skip("not running functional tests in --fast mode")
+    clientdir = request.config._tmpdirhandler.mktemp("liveserver2")
+    (p, url) = _liveserver(clientdir)
+    try:
+        yield url
+    finally:
+        p.terminate()
+        p.wait()
+
 
 @pytest.fixture
-def devpi(cmd_devpi, gen, url_of_liveserver):
-    user = gen.user()
+def devpi_username(gen):
+    return gen.user()
+
+
+@pytest.fixture
+def devpi(cmd_devpi, devpi_username, url_of_liveserver):
+    user = devpi_username
     cmd_devpi("use", url_of_liveserver.url, code=200)
     cmd_devpi("user", "-c", user, "password=123", "email=123")
     cmd_devpi("login", user, "--password", "123")
@@ -139,7 +205,7 @@ def initproj(request, tmpdir):
 
 @pytest.fixture
 def create_and_upload(request, devpi, initproj, Popen):
-    def upload(name, filedefs=None):
+    def upload(name, filedefs=None, opts=()):
         initproj(name, filedefs)
         # we need to patch .pypirc
         #with devpi.patched_pypirc as reponame:
@@ -151,7 +217,7 @@ def create_and_upload(request, devpi, initproj, Popen):
         #                   "-r", reponame])
         #    popen.communicate()
         #    assert popen.returncode == 0
-        devpi("upload")
+        devpi("upload", *opts)
     return upload
 
 
@@ -228,18 +294,20 @@ class Gen:
         return pkgname+"-1.0.tar.gz"
 
 
-def pytest_runtest_makereport(__multicall__, item, call):
+@pytest.hookimpl(hookwrapper=True)
+def pytest_runtest_makereport(item, call):
+    outcome = yield
     logfiles = getattr(item.config, "_extlogfiles", None)
     if logfiles is None:
         return
-    report = __multicall__.execute()
+    report = outcome.get_result()
     for name in sorted(logfiles):
         content = logfiles[name].read()
         if content:
             longrepr = getattr(report, "longrepr", None)
             if hasattr(longrepr, "addsection"):
                 longrepr.addsection("%s log" %name, content)
-    return report
+
 
 @pytest.fixture
 def ext_devpi(request, tmpdir, devpi):
@@ -265,6 +333,7 @@ def out_devpi(devpi):
                 devpi(*args, **kwargs)
             finally:
                 out, err = cap.reset()
+                del cap
         except:
             print_(out)
             print_(err)
@@ -275,8 +344,11 @@ def out_devpi(devpi):
     return out_devpi_func
 
 @pytest.fixture
-def cmd_devpi(tmpdir):
+def cmd_devpi(tmpdir, monkeypatch):
     """ execute devpi subcommand in-process (with fresh init) """
+    def ask_confirm(msg):
+        print("%s: yes" % msg)
+        return True
     clientdir = tmpdir.join("client")
     def run_devpi(*args, **kwargs):
         callargs = []
@@ -286,13 +358,23 @@ def cmd_devpi(tmpdir):
             callargs.append(str(arg))
         print_info("*** inline$ %s" % " ".join(callargs))
         hub, method = initmain(callargs)
+        monkeypatch.setattr(hub, "ask_confirm", ask_confirm)
+        expected = kwargs.get("code", None)
         try:
             method(hub, hub.args)
         except SystemExit as sysex:
             hub.sysex = sysex
-        expected = kwargs.get("code", None)
+            if expected == None or expected < 0 or expected >= 400:
+                # we expected an error or nothing, don't raise
+                pass
+            else:
+                raise
+        finally:
+            hub.close()
         if expected is not None:
-            if isinstance(expected, list):
+            if expected == -2:  # failed-to-start
+                assert hasattr(hub, "sysex")
+            elif isinstance(expected, list):
                 assert hub._last_http_stati == expected
             else:
                 if not isinstance(expected, tuple):
@@ -326,7 +408,6 @@ def runprocess(tmpdir, cmdargs):
 @pytest.fixture
 def mockhtml(monkeypatch):
     def mockhtml(cache, mockurl, content):
-        log.test("mock", mockurl)
         mockurl = cache._getsimpleurl(mockurl)
         old = cache.http.gethtml
         def newgethtml(url):
@@ -362,12 +443,13 @@ def create_venv(request, testdir, monkeypatch):
 
 
 @pytest.fixture
-def loghub(tmpdir, mock_http_api):
+def loghub(tmpdir):
     class args:
         debug = True
         clientdir = tmpdir.join("clientdir")
         yes = False
         verbose = False
+        settrusted = False
     out = py.io.TextIO()
     hub = Hub(args, file=out)
     def _getmatcher():
@@ -406,9 +488,13 @@ def mock_http_api(monkeypatch):
             self.called = []
             self._json_responses = {}
 
-        def __call__(self, method, url, kvdict=None, quiet=False, auth=None,
-                     fatal=True):
-            self.called.append((method, url, kvdict))
+        def __call__(self, method, url, kvdict=None, quiet=False,
+                     auth=None, basic_auth=None, cert=None,
+                     fatal=True, verify=None):
+            kwargs = dict(
+                kvdict=kvdict, quiet=quiet, auth=auth, basic_auth=basic_auth,
+                cert=cert, fatal=fatal)
+            self.called.append((method, url, kwargs))
             reply_data = self._json_responses.get(url)
             if reply_data is not None:
                 class R:
