@@ -1,9 +1,27 @@
-
+from io import BytesIO
+import py
 import pytest
-from py import std
+import requests
+import tarfile
 import time
 
 from test_devpi_server.functional import TestUserThings, TestIndexThings # noqa
+try:
+    from test_devpi_server.functional import TestMirrorIndexThings  # noqa
+except ImportError:
+    # when testing with older devpi-server
+    class TestMirrorIndexThings:
+        def test_mirror_things(self):
+            pytest.skip(
+                "Couldn't import TestMirrorIndexThings from devpi server tests.")
+try:
+    from test_devpi_server.functional import TestIndexPushThings  # noqa
+except ImportError:
+    # when testing with older devpi-server
+    class TestIndexPushThings:
+        def test_mirror_things(self):
+            pytest.skip(
+                "Couldn't import TestIndexPushThings from devpi server tests.")
 from test_devpi_server.functional import MappMixin
 
 @pytest.fixture
@@ -26,6 +44,22 @@ class Mapp(MappMixin):
         assert indexname
         return indexname
 
+    def makepkg(self, basename, content, name, version):
+        s = BytesIO()
+        pkg_info = '\n'.join([
+            "Metadata-Version: 1.1",
+            "Name: %s" % name,
+            "Version: %s" % version]).encode('utf-8')
+        tf = tarfile.open(basename, mode='w:gz', fileobj=s)
+        tinfo = tarfile.TarInfo('PKG-INFO')
+        tinfo.size = len(pkg_info)
+        tf.addfile(tinfo, BytesIO(pkg_info))
+        tinfo = tarfile.TarInfo('content')
+        tinfo.size = len(content)
+        tf.addfile(tinfo, BytesIO(content))
+        tf.close()
+        return s.getvalue()
+
     def cleanup(self):
         pw = getattr(self, "_rootpassword", None)
         if pw:
@@ -42,6 +76,9 @@ class Mapp(MappMixin):
     def logoff(self, code=None):
         self.devpi("logoff", code=code)
 
+    def logout(self, code=None):
+        self.devpi("logout", code=code)
+
     def use(self, indexname):
         assert indexname.count("/") == 1, indexname
         self.devpi("use", indexname)
@@ -59,12 +96,30 @@ class Mapp(MappMixin):
     def getjson(self, path, code=200):
         result = self.out_devpi("getjson", path, code=code)
         if code == 200:
-            return std.json.loads(result.stdout.str())
+            return py.std.json.loads(result.stdout.str())
 
     def getindexlist(self):
         result = self.out_devpi("index", "-l")
         return [x for x in result.outlines if x.strip()]
 
+    def getpkglist(self):
+        result = self.out_devpi("list")
+        return [x for x in result.outlines if x.strip()]
+
+    def getreleaseslist(self, name, code=200):
+        result = self.out_devpi("list", name, code=code)
+        return [x for x in result.outlines if x.strip()]
+
+    def downloadrelease(self, code, url):
+        r = requests.get(url)
+        if isinstance(code, tuple):
+            assert r.status_code in code
+        else:
+            assert r.status_code == code
+        if r.status_code < 300:
+            return r.content
+        return r.json()
+
     def change_password(self, user, password):
         auth = getattr(self, "auth", None)
         if auth is None or auth[0] != user and auth[0] != "root":
@@ -97,8 +152,10 @@ class Mapp(MappMixin):
             for name, val in indexconfig.items():
                 if name == "bases":
                     params.append("%s=%s" % (name, ",".join(val)))
-                if name == "volatile":
+                elif name == "volatile":
                     params.append("%s=%s" % (name, bool(val)))
+                else:
+                    params.append("%s=%s" % (name, val))
         return params
 
     def create_index(self, indexname, indexconfig=None, code=200):
@@ -123,25 +180,32 @@ class Mapp(MappMixin):
         self.devpi("index", indexname, "acl_upload=%s" % acls, code=200)
 
     def set_custom_data(self, data, indexname=None):
-        if indexname is None:
-            self.devpi("index", "custom_data=%s" % data, code=200)
-        else:
-            self.devpi("index", indexname,
-                       "custom_data=%s" % data, code=200)
+        return self.set_key_value("custom_data", data, indexname=indexname)
 
-    def set_uploadtrigger_jenkins(self, url, indexname=None):
+    def set_key_value(self, key, value, indexname=None):
         if indexname is None:
-            self.devpi("index", "uploadtrigger_jenkins=%s" % url, code=200)
+            self.devpi("index", "%s=%s" % (key, value), code=200)
         else:
             self.devpi("index", indexname,
-                       "uploadtrigger_jenkins=%s" % url, code=200)
+                       "%s=%s" % (key, value), code=200)
 
-    def set_pypi_whitelist(self, whitelist, indexname=None):
+    def set_uploadtrigger_jenkins(self, *args, **kwargs):
+        # called when we run client tests against server-2.1
+        pytest.skip("jenkins functionality moved out to pytest-jenkins")
+
+    def set_indexconfig_option(self, key, value, indexname=None):
         if indexname is None:
-            self.devpi("index", "pypi_whitelist=%s" % whitelist, code=200)
+            self.devpi("index", "%s=%s" % (key, value), code=200)
         else:
             self.devpi("index", indexname,
-                       "pypi_whitelist=%s" % whitelist, code=200)
+                       "%s=%s" % (key, value), code=200)
+
+    def set_mirror_whitelist(self, whitelist, indexname=None):
+        if indexname is None:
+            self.devpi("index", "mirror_whitelist=%s" % whitelist, code=200)
+        else:
+            self.devpi("index", indexname,
+                       "mirror_whitelist=%s" % whitelist, code=200)
 
     def get_acl(self, code=200, indexname=None):
         indexname = self._getindexname(indexname)
@@ -153,14 +217,24 @@ class Mapp(MappMixin):
                 return [x for x in parts[1].split(",") if x]
         return  []
 
-    def get_pypi_whitelist(self, code=200, indexname=None):
+    def get_mirror_whitelist(self, code=200, indexname=None):
         indexname = self._getindexname(indexname)
         result = self.out_devpi("index", indexname)
         for line in result.outlines:
             line = line.strip()
-            parts = line.split("pypi_whitelist=", 1)
+            parts = line.split("mirror_whitelist=", 1)
             return parts[1].split(",")
 
+    def upload_file_pypi(self, basename, content,
+                         name=None, version=None):
+        assert py.builtin._isbytes(content)
+        pkg = self.tmpdir.join(basename)
+        pkg.write_binary(content)
+        self.devpi('upload', pkg.strpath)
+
+    def push(self, name, version, index, indexname=None, code=200):
+        self.devpi('push', '%s==%s' % (name, version), index, code=code)
+
     def create_project(self, projectname, code=201, indexname=None):
         pytest.xfail(reason="no way to create project via command line yet")
 
@@ -170,10 +244,42 @@ def test_logoff(mapp):
     mapp.logoff()
     mapp.logoff()
 
+
+def test_logout(mapp):
+    mapp.login()
+    mapp.logout()
+    mapp.logout()
+
+
 def test_getjson(out_devpi):
     result = out_devpi("getjson", "/", "-v")
     assert "X-DEVPI-API-VERSION" in result.stdout.str()
 
+
+def test_switch_preserves_auth(out_devpi, url_of_liveserver, url_of_liveserver2):
+    import re
+    result1 = out_devpi("use", url_of_liveserver)
+    (url1, user1) = re.search(
+        '(https?://.+?)\s+\(logged in as (.+?)\)', result1.stdout.str()).groups()
+    result2 = out_devpi("use", url_of_liveserver2)
+    url2 = re.search(
+        '(https?://.+?)\s+\(not logged in\)', result2.stdout.str()).group(1)
+    assert url2 != url1
+    out_devpi("user", "-c", user1, "password=123", "email=123")
+    out_devpi("login", user1, "--password", "123")
+    out_devpi("index", "-c", "dev")
+    result3 = out_devpi("use", "dev")
+    (url3, user3) = re.search(
+        '(https?://.+?)\s+\(logged in as (.+?)\)', result3.stdout.str()).groups()
+    assert user3 == user1
+    assert url3.startswith(url2)
+    result4 = out_devpi("use", url_of_liveserver)
+    (url4, user4) = re.search(
+        '(https?://.+?)\s+\(logged in as (.+?)\)', result4.stdout.str()).groups()
+    assert user4 == user1
+    assert url4 == url1
+
+
 class TestUserManagement:
     """ This class tests the user sub command of devpi
     """
