@@ -1,4 +1,5 @@
 import py
+import pytest
 import json
 from devpi.main import Hub, check_output
 from devpi.push import parse_target, PyPIPush, DevpiPush
@@ -10,6 +11,7 @@ def runproc(cmd):
 def test_parse_target_devpi(loghub):
     class args:
         target = "user/name"
+        index = None
     res = parse_target(loghub, args)
     assert isinstance(res, DevpiPush)
 
@@ -27,6 +29,7 @@ def test_parse_target_pypi(tmpdir, loghub):
     class args:
         target = "pypi:whatever"
         pypirc = str(p)
+        index = None
     res = parse_target(loghub, args)
     assert isinstance(res, PyPIPush)
     assert res.user == "test"
@@ -47,6 +50,7 @@ def test_parse_target_pypi_default_repository(tmpdir, loghub):
     class args:
         target = "pypi:whatever"
         pypirc = str(p)
+        index = None
     res = parse_target(loghub, args)
     assert isinstance(res, PyPIPush)
     assert res.user == "test"
@@ -57,6 +61,7 @@ def test_parse_target_pypi_default_repository(tmpdir, loghub):
 def test_push_devpi(loghub, monkeypatch, mock_http_api):
     class args:
         target = "user/name"
+        index = None
     pusher = parse_target(loghub, args)
     mock_http_api.set(loghub.current.index, 200, result={})
     pusher.execute(loghub, "pytest", "2.3.5")
@@ -65,10 +70,23 @@ def test_push_devpi(loghub, monkeypatch, mock_http_api):
     # loghub.http_api.assert_called_once_with(
     #            "push", loghub.current.index, kvdict=req)
 
-def test_main_push_pypi(monkeypatch, tmpdir):
+
+def test_push_devpi_index_option(loghub, monkeypatch, mock_http_api):
+    class args:
+        target = "user/name"
+        index = "src/dev"
+    pusher = parse_target(loghub, args)
+    mock_http_api.set("src/dev", 200, result={})
+    pusher.execute(loghub, "pytest", "2.3.5")
+    dict(name="pytest", version="2.3.5", targetindex="user/name")
+    assert len(mock_http_api.called) == 1
+
+
+@pytest.mark.parametrize("spec", ("pkg==1.0", "pkg-1.0"))
+def test_main_push_pypi(monkeypatch, tmpdir, spec):
     from devpi.push import main
     l = []
-    def mypost(method, url, data, headers, auth=None, cert=None):
+    def mypost(method, url, data, headers, auth=None, cert=None, verify=None):
         l.append((method, url, data))
         class r:
             status_code = 201
@@ -85,6 +103,7 @@ def test_main_push_pypi(monkeypatch, tmpdir):
     class args:
         clientdir = tmpdir.join("client")
         debug = False
+        index = None
     hub = Hub(args)
     monkeypatch.setattr(hub.http, "request", mypost)
     hub.current.reconfigure(dict(index="/some/index"))
@@ -101,7 +120,8 @@ def test_main_push_pypi(monkeypatch, tmpdir):
     class args:
         pypirc = str(p)
         target = "pypi:whatever"
-        nameversion = "pkg-1.0"
+        pkgspec = spec
+        index = None
 
     main(hub, args)
     assert len(l) == 1
@@ -114,6 +134,56 @@ def test_main_push_pypi(monkeypatch, tmpdir):
     assert req["username"] == "test"
     assert req["password"] == "testp"
 
+
+def test_fail_push(monkeypatch, tmpdir):
+    from devpi.push import main
+    l = []
+    def mypost(method, url, data, headers, auth=None, cert=None, verify=None):
+        l.append((method, url, data))
+        class r:
+            status_code = 500
+            reason = "Internal Server Error"
+            content = json.dumps(dict(type="actionlog", status=201,
+                result=[("500", "Internal Server Error", "Internal Server Error")]
+            ))
+            headers = {"content-type": "application/json"}
+            _json = json.loads(content)
+            class request:
+                method = ''
+        r.url = url
+        r.request.method = method
+
+        return r
+
+    class args:
+        clientdir = tmpdir.join("client")
+        debug = False
+        index = None
+    hub = Hub(args)
+    monkeypatch.setattr(hub.http, "request", mypost)
+    hub.current.reconfigure(dict(index="/some/index"))
+    p = tmpdir.join("pypirc")
+    p.write(py.std.textwrap.dedent("""
+        [distutils]
+        index-servers = whatever
+
+        [whatever]
+        repository: http://anotherserver
+        username: test
+        password: testp
+    """))
+    class args:
+        pypirc = str(p)
+        target = "pypi:whatever"
+        pkgspec = "pkg==1.0"
+        index = None
+
+    try:
+        main(hub, args)
+    except SystemExit as e:
+        assert e.code==1
+
+
 class TestPush:
     def test_help(self, ext_devpi):
         result = ext_devpi("push", "-h")
